
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  DIR.C
*   DESCRIP  :  FENRIS VFS Dir Module for Linux
*   DATE     :  December 5, 1998
*
*
***************************************************************************/

#include "globals.h"

#if (LINUX_22)

int nwfs_dentry_validate(struct dentry *dentry, int flags)
{
#if (VFS_DEBUG)
    NWFSPrint("dentry validate %s\n", dentry->d_name.name);
#endif
    return 1;
}

struct dentry_operations nwfs_dentry_operations =
{
    nwfs_dentry_validate,	// d_revalidate(struct dentry *, int)
    NULL,			// d_hash
    NULL,			// d_compare
    NULL, 		  	// d_delete(struct dentry *)
    NULL,			// d_release(struct dentry *)
    NULL,			// d_iput(struct dentry *, struct inode *)
};

ssize_t nwfs_dir_read(struct file *filp, char *buf,
		      size_t count, loff_t *ppos)
{
    return -EISDIR;
}

int nwfs_dir_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
    register ULONG i, count = 0;
    struct inode *inode = filp->f_dentry->d_inode;
    register VOLUME *volume = (VOLUME *) inode->i_sb->u.generic_sbp;
    register HASH *hash = 0;

    if (!inode || !S_ISDIR(inode->i_mode))
       return -ENOTDIR;

    i = filp->f_pos;
    switch (i)
    {
       case 0:
	  if (filldir(dirent, ".",  1, filp->f_pos, inode->i_ino) < 0)
	     goto readdir_end;
	  filp->f_pos++;
          count++;

       case 1:
	  if (filldir(dirent, "..", 2, filp->f_pos,
                      filp->f_dentry->d_parent->d_inode->i_ino) < 0)
	     goto readdir_end;
	  filp->f_pos++;
          count++;
    }

    while (1)
    {
       hash = get_subdirectory_record(volume, filp->f_pos, inode->i_ino);
       if (!hash)
          break;

       // provide name and root name space (MSDOS) record DirNo
       if (filldir(dirent, hash->Name, hash->NameLength,
			 filp->f_pos, hash->Root) < 0)

	  goto readdir_end;
       filp->f_pos++;
       count++;
    }

readdir_end:;

#if (VFS_DEBUG)
    NWFSPrint("readdir count=%d pos1-%d pos2-%d\n", (int)count,
             (int)i, (int)filp->f_pos);
#endif

    return count;
}

int nwfs_dir_ioctl(struct inode * inode, struct file * filp,
		  unsigned int cmd, unsigned long arg)
{

#if (VFS_DEBUG)
    NWFSPrint("ioctl cmd-%d\n", (int)cmd);
#endif

    switch (cmd)
    {
       default:
	  return -EINVAL;
    }
    return 0;
}

struct dentry *nwfs_dir_lookup(struct inode *dir, struct dentry *dentry)
{
    register VOLUME *volume = (VOLUME *) dir->i_sb->u.generic_sbp;
    register ino_t ino;
    struct inode **result = &(dentry->d_inode);

#if (VFS_DEBUG)
    NWFSPrint("lookup %s\n", dentry->d_name.name);
#endif

    *result = NULL;
    if (!dir)
       return ERR_PTR(-ENOENT);

    if (!S_ISDIR(dir->i_mode))
       return ERR_PTR(-ENOENT);

    if ((dentry->d_name.len == 0) || ((dentry->d_name.name[0] == '.') &&
	(dentry->d_name.len == 1)))
    {
       *result = dir;
       return 0;
    }

    ino = get_directory_number(volume,
			       &dentry->d_name.name[0],
			       dentry->d_name.len,
			       dir->i_ino);
    if (ino)
    {
       *result = iget(dir->i_sb, ino);
       if (!(*result))
       {
          d_add(dentry, NULL); // assume create or mkdir being attempted
          dentry->d_time = 0;
//          dentry->d_op = &nwfs_dentry_operations;
          return 0;
       }
       d_add(dentry, *result);
       dentry->d_time = 0;
//       dentry->d_op = &nwfs_dentry_operations;
       return 0;
    }
    else
    {
       d_add(dentry, NULL); // assume create or mkdir being attempted
       dentry->d_time = 0;
//       dentry->d_op = &nwfs_dentry_operations;
       return 0;
    }
}

struct file_operations nwfs_dir_operations = {
    NULL,			// lseek
    nwfs_dir_read,		// read
    NULL,			// write
    nwfs_dir_readdir,		// readdir
    NULL,			// select v2.0.x/poll v2.1.x - default
    nwfs_dir_ioctl,		// ioctl
    NULL,			// mmap
    NULL,			// open
    NULL,			// flush
    NULL,			// release
    NULL,		        // fsync
    NULL,
    NULL,
    NULL
};

struct inode_operations nwfs_dir_inode_operations =
{
    &nwfs_dir_operations,	// file operations
    nwfs_create,		// create
    nwfs_dir_lookup,		// lookup
    nwfs_link,             	// link
    nwfs_unlink,		// unlink
    nwfs_symlink,	        // symlink
    nwfs_mkdir,			// mkdir
    nwfs_rmdir,			// rmdir
    nwfs_mknod,   		// mknod
    nwfs_rename,		// rename
    NULL,			// readlink
    NULL,			// follow_link
    NULL,			// read_page
    NULL,			// writepage
    NULL,			// bmap
    NULL,			// truncate
    NULL,			// permission
    NULL,			// smap
    NULL,			// update page
    NULL, 		        // revalidate inode
};

#endif


#if (LINUX_20)

int nwfs_dir_read(struct inode *inode, struct file *filp,
		    char *buf, int count)
{
    return -EISDIR;
}

int nwfs_dir_readdir(struct inode *inode, struct file *filp,
		       void *dirent, filldir_t filldir)
{
    register ULONG i, count = 0;
    register VOLUME *volume = (VOLUME *) inode->i_sb->u.generic_sbp;
    register HASH *hash = 0;

    if (!inode || !S_ISDIR(inode->i_mode))
       return -ENOTDIR;

    i = filp->f_pos;
    switch (i)
    {
       case 0:
	  if (filldir(dirent, ".",  1, filp->f_pos, inode->i_ino) < 0)
	     return count;
	  filp->f_pos++;
          count++;

       case 1:
	  if (filldir(dirent, "..", 2, filp->f_pos,
		   get_parent_directory_number(volume, inode->i_ino)) < 0)
	     return count;
	  filp->f_pos++;
          count++;
    }

    while (1)
    {
       hash = get_subdirectory_record(volume, filp->f_pos, inode->i_ino);
       if (!hash)
          break;

       // provide name and root name space (MSDOS) record DirNo
       if (filldir(dirent, hash->Name, hash->NameLength, filp->f_pos,
                   hash->Root) < 0)
	  return count;
       filp->f_pos++;
       count++;
    }
    return count;
}

int nwfs_dir_lookup(struct inode *dir, const char *name, int len,
		      struct inode **result)
{
	register VOLUME *volume = (VOLUME *) dir->i_sb->u.generic_sbp;
	register ino_t ino;
	register struct super_block *sb;

	*result = 0;
	if (!dir)
	{
	   return -ENOENT;
	}

	if (!S_ISDIR(dir->i_mode))
	{
	   iput(dir);
	   return -ENOENT;
	}

	if ((len == 0) || ((name[0] == '.') && (len == 1)))
	{
	   *result = dir;
	   return 0;
	}

	if ((name[0] == '.') &&( name[1] == '.') && (len == 2))
	{
	   sb = dir->i_sb;
	   ino = get_parent_directory_number(volume, dir->i_ino);
	   iput(dir);

	   *result = iget(sb, ino);
	   if (!*result)
	      return -EACCES;
	   return 0;
	}

	ino = get_directory_number(volume, name, len, dir->i_ino);
	if (!ino)
	{
	   iput(dir);
	   return -ENOENT;
	}

	*result = iget(dir->i_sb, ino);
	if (!*result)
	{
	   iput(dir);
	   return -EACCES;
	}
	iput(dir);
	return 0;

}

struct file_operations nwfs_dir_operations =
{
	NULL,			// lseek
	nwfs_dir_read,	// read
	NULL,			// write
	nwfs_dir_readdir,	// readdir
	NULL,			// select
	NULL,			// ioctl
	NULL,			// mmap
	NULL,			// open
	NULL,			// release
	NULL,			// fsync
	NULL,			// fasync
	NULL,			// check_media_change
	NULL			// revalidate
};

struct inode_operations nwfs_dir_inode_operations =
{
	&nwfs_dir_operations,	// file operations
	nwfs_create,		// create
	nwfs_dir_lookup,	// lookup
	nwfs_link,             	// link
	nwfs_unlink,		// unlink
	nwfs_symlink,	        // symlink
	nwfs_mkdir,		// mkdir
	nwfs_rmdir,		// rmdir
	nwfs_mknod,   		// mknod
	nwfs_rename,		// rename
        NULL,			// readlink
	NULL,			// follow_link
	NULL,			// read_page
	NULL,			// writepage
	NULL,			// bmap
	NULL,			// truncate
	NULL,			// permission
	NULL			// smap
};

#endif
