
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  NWFS.H
*   DESCRIP  :  FENRIS Volume and System On Disk Structures
*   DATE     :  November 1, 1998
*
*
***************************************************************************/

#ifndef _NWFS_
#define _NWFS_

#define NETWARE_286_ID     0x64
#define NETWARE_386_ID     0x65
#define PARTITION_ID     0xAA55
#define NW3X_PARTITION        1
#define NW4X_PARTITION        0

#define MAX_DISKS            64
#define MAX_PARTITIONS      MAX_DISKS * 4
#define MAX_VOLUMES         256
#define MAX_VOLUME_ENTRIES    8
#define MAX_SEGMENTS         32
#define MAX_MIRRORS           8
#define MAX_NAMESPACES       10  // 12 in 4.x, 10 in 3.x
#define MAX_SUBALLOC_NODES    5
#define MAX_SUBALLOC_SPAN     2
#define MAX_HOTFIX_BLOCKS    30
#define MAX_FLUSH           500
#define LRU_FACTOR            2
#define LRU_AGE              10

#define FOUR_K_BLOCK          3
#define EIGHT_K_BLOCK         4
#define SIXTEEN_K_BLOCK       5
#define THIRTYTWO_K_BLOCK     6
#define SIXTYFOUR_K_BLOCK     7

// these defines describe the location of the Hot Fix and Mirroring Tables

#define MULTI_TRACK_COUNT     4

#define VOLUME_TABLE_LOCATION_0   0x20
#define VOLUME_TABLE_LOCATION_1   0x40
#define VOLUME_TABLE_LOCATION_2   0x60
#define VOLUME_TABLE_LOCATION_3   0x80

#define HOTFIX_LOCATION_0   0x20
#define HOTFIX_LOCATION_1   0x40
#define HOTFIX_LOCATION_2   0x60
#define HOTFIX_LOCATION_3   0x80

#define MIRROR_LOCATION_0   0x21
#define MIRROR_LOCATION_1   0x41
#define MIRROR_LOCATION_2   0x61
#define MIRROR_LOCATION_3   0x81

#define MIN_FAT_LRU_BLOCKS        0
#define MAX_FAT_LRU_BLOCKS        100
#define MIN_DIRECTORY_LRU_BLOCKS  10
#define MAX_DIRECTORY_LRU_BLOCKS  500
#define MIN_VOLUME_LRU_BLOCKS     10
#define MAX_VOLUME_LRU_BLOCKS     1000
#define MAX_WORKSPACE_BUFFERS     32
#define MAX_BUFFER_HEADS          1000

#if (DOS_UTIL)
#undef  MAX_VOLUME_LRU_BLOCKS
#define MAX_VOLUME_LRU_BLOCKS     500
#endif

#if (WINDOWS_NT | WINDOWS_NT_RO)
#undef  MAX_VOLUME_LRU_BLOCKS
#define MAX_VOLUME_LRU_BLOCKS     1000
#endif

#if (LINUX_20 | LINUX_22)
#undef  MAX_VOLUME_LRU_BLOCKS
#define MAX_VOLUME_LRU_BLOCKS     2000
#endif

#define IO_BLOCK_SIZE             4096
#define BIT_BLOCK_SIZE            4096
#define HOTFIX_BLOCK_SIZE         4096
#define SUBALLOC_BLOCK_SIZE       512

#define HOTFIX_SIGNATURE "HOTFIX00"
#define MIRROR_SIGNATURE "MIRROR00"
#define VOLUME_SIGNATURE "NetWare Volumes"

#define USER_ADDRESS_SPACE        0
#define KERNEL_ADDRESS_SPACE      1

//
//  memory tracking labels
//

#define HOTFIX_TAG         (ULONG)"HTFX"
#define HOTFIX_DATA_TAG    (ULONG)"HFDT"
#define HOTFIX_TABLE_TAG   (ULONG)"HFTB"
#define MIRROR_TAG         (ULONG)"MIRR"
#define MIRROR_DATA_TAG    (ULONG)"MIRD"
#define HASH_TAG           (ULONG)"HASH"
#define EXT_HASH_TAG       (ULONG)"EHSH"
#define BITBLOCK_TAG       (ULONG)"BITB"
#define BITBUFFER_TAG      (ULONG)"BBUF"
#define NWPART_TAG         (ULONG)"PART"
#define NWDISK_TAG         (ULONG)"DISK"
#define DISKBUF_TAG        (ULONG)"DSKB"
#define TABLE_TAG          (ULONG)"TBLE"
#define FAT_TAG            (ULONG)"FATM"
#define FATLRU_TAG         (ULONG)"FATL"
#define FATHASH_TAG        (ULONG)"FATH"
#define DIRLRU_TAG         (ULONG)"DIRL"
#define DIRHASH_TAG        (ULONG)"DIRH"
#define DIR_TAG            (ULONG)"DIRM"
#define DIR_BLOCK_TAG      (ULONG)"DIRB"
#define DIR_BLOCKHASH_TAG  (ULONG)"DRBH"
#define EXT_BLOCK_TAG      (ULONG)"EXTB"
#define EXT_BLOCKHASH_TAG  (ULONG)"EXBH"
#define ASSN_BLOCK_TAG     (ULONG)"ASNB"
#define ASSN_BLOCKHASH_TAG (ULONG)"ASBH"
#define BLOCK_TAG          (ULONG)"BLKM"
#define CREATE_TAG         (ULONG)"CRET"
#define TRUSTEE_TAG        (ULONG)"TRST"
#define USER_TAG           (ULONG)"USER"
#define NWDIR_TAG          (ULONG)"NWDT"
#define NWEXT_TAG          (ULONG)"NWET"
#define FAT_WORKSPACE_TAG  (ULONG)"WRKF"
#define DIR_WORKSPACE_TAG  (ULONG)"WRKD"
#define EXT_WORKSPACE_TAG  (ULONG)"WRKE"
#define CRT_WORKSPACE_TAG  (ULONG)"WRKC"
#define SA_WORKSPACE_TAG   (ULONG)"WKSA"
#define TRUNC_WORK_TAG     (ULONG)"WKST"
#define SECTOR_TAG         (ULONG)"SECT"
#define CLUSTER_TAG        (ULONG)"CLST"
#define VOLUME_TAG         (ULONG)"VOLS"
#define VOLUME_WS_TAG      (ULONG)"VOLW"
#define VOLUME_DATA_TAG    (ULONG)"VOLD"
#define LRU_TAG            (ULONG)"LRUS"
#define LRU_HASH_TAG       (ULONG)"LRUH"
#define LRU_BUFFER_TAG     (ULONG)"LRUB"
#define ALIGN_BUF_TAG      (ULONG)"ALGN"
#define NAM_HASH_TAG       (ULONG)"NAMH"
#define DNM_HASH_TAG       (ULONG)"DNMH"
#define PAR_HASH_TAG       (ULONG)"PARH"
#define EXD_HASH_TAG       (ULONG)"EXDH"
#define TRS_HASH_TAG       (ULONG)"TRSH"
#define DEL_HASH_TAG       (ULONG)"DELH"
#define QUO_HASH_TAG       (ULONG)"QUOH"
#define SUBALLOC_HEAD_TAG  (ULONG)"SUBH"
#define PAGED_FAT_BUF      (ULONG)"PGFT"
#define PAGED_WORK_BUF     (ULONG)"PGWK"
#define PAGED_PART_BUF     (ULONG)"PGPT"
#define PAGED_VOL_BUF      (ULONG)"PGVL"
#define PAGED_SECT_BUF     (ULONG)"PGSC"
#define DIR_WORKSPACE      (ULONG)"DIRW"
#define SEMAPHORE_TAG      (ULONG)"SEMA"
#define NAME_STORAGE_TAG   (ULONG)"NAME"
#define NWVP_TAG           (ULONG)"NWVP"
#define BH_TAG             (ULONG)"BHBH"

#if (WINDOWS_NT | WINDOWS_NT_RO)
#define NT_POOL_TAG        'sFwN'
#endif

//
//  Hotfix and Mirroring Tables occupy the first sections
//  of a Netware 386 Partition.  There are four copies of these
//  tables.  Netware treats all on-disk structures as zero relative
//  from the beginning of any segment of the disk.  Following
//  these tables are the Volume Segment Tables that define
//  specific information about a volume.  There are four copies of these
//  tables as well.  All of these copies are aligned on track boundries
//  for fault-tolerance so in the event a track fails, at least one good
//  copy of these tables is preserved.
//
//  The Volume Tables contain all the information about volume
//  segments on that drive, and volumes that occupy the partition.
//  Also in these tables are stored the logical starting block for
//  the volume FAT and the first Directory Entry.  The Netware directory
//  is actually treated just like a file, and is mapped via the FAT
//  table.  This is different that the way MS-DOS implements their
//  directory, which chains directory blocks and stores an LBA in each
//  directory object.
//
//  Volumes that have more than one volume segment require that the
//  host operating system read all the disks present in the system
//  and their corresponding VolumeTables to locate the segments for
//  a particular Netware Volume.  If any segments or disks are missing
//  then it is impossible to properly mount a Netware Volume unless a
//  mirror is present.
//
//  The Netware FAT (File Allocation Table) begins at logical
//  block 0 on a logical Netware volume.  There are always two
//  copies of the FAT per logical volume.  The first field of the FAT
//  points to the file index.  Netware allows files to have "holes"
//  to preserve disk space.  For example, if someone creates a file
//  that is 4MB, then only writes two blocks, one a 0 and another at 16K,
//  the index field will point to the actual block index within the
//  logical file (i.e. FAT enties would be 0, 4).  The second field is
//  the continuation index into the FAT to the next entry for the file.
//  The Netware FAT is identical to a DOS FAT in that it is a true
//  one-to-one mapping block-to-fat entry.  If someone tries to read an
//  offset within a file that overlaps a "hole", zeros are returned
//  indicating that no data has actually been written there.
//

#if (LINUX | LINUX_UTIL | DOS_UTIL)

/* Boot sector info (62 byte structure) */

struct BOOT_SECTOR
{
    BYTE Jmp[3]         __attribute__ ((packed));
    BYTE OEMname[8]     __attribute__ ((packed));
    WORD bps            __attribute__ ((packed));
    BYTE SecPerClstr    __attribute__ ((packed));
    WORD ResSectors     __attribute__ ((packed));
    BYTE FATs           __attribute__ ((packed));
    WORD RootDirEnts    __attribute__ ((packed));
    WORD Sectors        __attribute__ ((packed));
    BYTE Media          __attribute__ ((packed));
    WORD SecPerFAT      __attribute__ ((packed));
    WORD SecPerTrack    __attribute__ ((packed));
    WORD Heads          __attribute__ ((packed));
    ULONG HiddenSecs    __attribute__ ((packed));
    ULONG HugeSecs      __attribute__ ((packed));
    BYTE DriveNum       __attribute__ ((packed));
    BYTE Rsvd1          __attribute__ ((packed));
    BYTE BootSig        __attribute__ ((packed));
    ULONG VolID         __attribute__ ((packed));
    BYTE VolLabel[11]   __attribute__ ((packed));
    BYTE FileSysType[8] __attribute__ ((packed));       /* 62 bytes */
};

/* Partition Table Entry info. 16 bytes */

struct PartitionTableEntry
{
    BYTE fBootable      __attribute__ ((packed));
    BYTE HeadStart      __attribute__ ((packed));
    BYTE SecStart       __attribute__ ((packed));
    BYTE CylStart       __attribute__ ((packed));
    BYTE SysFlag        __attribute__ ((packed));
    BYTE HeadEnd        __attribute__ ((packed));
    BYTE SecEnd         __attribute__ ((packed));
    BYTE CylEnd         __attribute__ ((packed));
    ULONG StartLBA      __attribute__ ((packed));
    ULONG nSectorsTotal __attribute__ ((packed));
};

typedef struct _VOLUME_TABLE_ENTRY
{
   BYTE VolumeName[16]        __attribute__ ((packed));  // length preceded string format
   ULONG LastVolumeSegment    __attribute__ ((packed));
   ULONG VolumeSignature      __attribute__ ((packed));
   ULONG VolumeRoot           __attribute__ ((packed));
   ULONG SegmentSectors       __attribute__ ((packed));
   ULONG VolumeClusters       __attribute__ ((packed));
   ULONG SegmentClusterStart  __attribute__ ((packed)); // Logical Block Segment Start
   ULONG FirstFAT             __attribute__ ((packed));            // FAT index for 1st FAT Table
   ULONG SecondFAT            __attribute__ ((packed));           // FAT index for 2nd FAT Table
   ULONG FirstDirectory       __attribute__ ((packed));      // FAT index for 1st Directory Block
   ULONG SecondDirectory      __attribute__ ((packed));     // FAT index for 2nd Directory Block
   ULONG Padding              __attribute__ ((packed));
} VOLUME_TABLE_ENTRY;

typedef struct _VOLUME_TABLE
{
   BYTE VolumeTableSignature[16]   __attribute__ ((packed));  // length preceded string here
   ULONG NumberOfTableEntries      __attribute__ ((packed));
   ULONG Reserved[3]               __attribute__ ((packed));
   VOLUME_TABLE_ENTRY VolumeEntry[MAX_VOLUME_ENTRIES]  __attribute__ ((packed));
} VOLUME_TABLE;

typedef struct _HOT_FIX_BLOCK_TABLE
{
   ULONG HotFix1    __attribute__ ((packed));    // these tables are sector based offsets from
   ULONG BadBlock1  __attribute__ ((packed));
   ULONG HotFix2    __attribute__ ((packed));    // the beginning of the partition
   ULONG BadBlock2  __attribute__ ((packed));
} HOTFIX_BLOCK_TABLE;

typedef struct _HOTFIX {
   BYTE HotFixStamp[8]      __attribute__ ((packed));
   ULONG PartitionID        __attribute__ ((packed));      // Novell HDK/SDK says .....
   ULONG HotFixFlags        __attribute__ ((packed));      // (WORD / WORD - data bits/synch number)
   ULONG HotFixDateStamp    __attribute__ ((packed));  // (B/B/B/B BSize/BUMask/BShift/BMask)
   ULONG HotFixTotalSectors __attribute__ ((packed));
   ULONG HotFixSize         __attribute__ ((packed));
   HOTFIX_BLOCK_TABLE HotFixTable[MAX_HOTFIX_BLOCKS] __attribute__ ((packed));
} HOTFIX;

typedef struct _MIRROR
{
   BYTE MirrorStamp[8]      __attribute__ ((packed));
   ULONG PartitionID        __attribute__ ((packed));      // Novell's website says ....
   ULONG MirrorFlags        __attribute__ ((packed));      // (B/B/W  DataBadBits/MirrorActive/SyncNumber)
   ULONG Reserved1          __attribute__ ((packed));        // (B/B/B/B  BSize/BUMask/BShift/BMask)
   ULONG MirrorStatus       __attribute__ ((packed));     // (B/B/W DeviceBadBits/DeviceActive/SynchNumber)
   ULONG MirrorTotalSectors __attribute__ ((packed));
   ULONG MirrorGroupID      __attribute__ ((packed));
   ULONG MirrorMemberID[MAX_MIRRORS] __attribute__ ((packed));
} MIRROR;

typedef struct _FAT_ENTRY
{
   long FATIndex      __attribute__ ((packed));
   long FATCluster    __attribute__ ((packed));
} FAT_ENTRY;

typedef struct _SUBALLOC_DIR
{
    ULONG  Flag                 __attribute__ ((packed));
    ULONG  Reserved1            __attribute__ ((packed));
    BYTE   ID                   __attribute__ ((packed));
    BYTE   SequenceNumber       __attribute__ ((packed));
    BYTE   Reserved2[2]         __attribute__ ((packed));
    ULONG  SubAllocationList    __attribute__ ((packed));
    ULONG  StartingFATChain[28] __attribute__ ((packed));
} SUBALLOC_DIR;

#endif


#if (WINDOWS_NT | WINDOWS_NT_RO | WINDOWS_NT_UTIL)

/* Boot sector info (62 byte structure) */

struct BOOT_SECTOR
{
    BYTE Jmp[3];
    BYTE OEMname[8];
    WORD bps;
    BYTE SecPerClstr;
    WORD ResSectors;
    BYTE FATs;
    WORD RootDirEnts;
    WORD Sectors;
    BYTE Media;
    WORD SecPerFAT;
    WORD SecPerTrack;
    WORD Heads;
    ULONG HiddenSecs;
    ULONG HugeSecs;
    BYTE DriveNum;
    BYTE Rsvd1;
    BYTE BootSig;
    ULONG VolID;
    BYTE VolLabel[11];
    BYTE FileSysType[8];        /* 62 bytes */
};

/* Partition Table Entry info. 16 bytes */

struct PartitionTableEntry
{
    BYTE fBootable;
    BYTE HeadStart;
    BYTE SecStart;
    BYTE CylStart;
    BYTE SysFlag;
    BYTE HeadEnd;
    BYTE SecEnd;
    BYTE CylEnd;
    ULONG StartLBA;
    ULONG nSectorsTotal;
};

typedef struct _VOLUME_TABLE_ENTRY
{
   BYTE VolumeName[16];  // length preceded string format
   ULONG LastVolumeSegment;
   ULONG VolumeSignature;
   ULONG VolumeRoot;
   ULONG SegmentSectors;
   ULONG VolumeClusters;
   ULONG SegmentClusterStart; // Logical Block Segment Start
   ULONG FirstFAT;            // FAT index for 1st FAT Table
   ULONG SecondFAT;           // FAT index for 2nd FAT Table
   ULONG FirstDirectory;      // FAT index for 1st Directory Block
   ULONG SecondDirectory;     // FAT index for 2nd Directory Block
   ULONG Padding;
} VOLUME_TABLE_ENTRY;

typedef struct _VOLUME_TABLE
{
   BYTE VolumeTableSignature[16];  // length preceded string here
   ULONG NumberOfTableEntries;
   ULONG Reserved[3];
   VOLUME_TABLE_ENTRY VolumeEntry[MAX_VOLUME_ENTRIES];
} VOLUME_TABLE;

typedef struct _HOT_FIX_BLOCK_TABLE
{
   ULONG HotFix1;    // these tables are sector based offsets from
   ULONG BadBlock1;
   ULONG HotFix2;    // the beginning of the partition
   ULONG BadBlock2;
} HOTFIX_BLOCK_TABLE;

typedef struct _HOTFIX {
   BYTE HotFixStamp[8];
   ULONG PartitionID;      // Novell HDK/SDK says .....
   ULONG HotFixFlags;      // (WORD / WORD - data bits/synch number)
   ULONG HotFixDateStamp;  // (B/B/B/B BSize/BUMask/BShift/BMask)
   ULONG HotFixTotalSectors;
   ULONG HotFixSize;
   HOTFIX_BLOCK_TABLE HotFixTable[MAX_HOTFIX_BLOCKS];
} HOTFIX;

typedef struct _MIRROR
{
   BYTE MirrorStamp[8];
   ULONG PartitionID;      // Novell's website says ....
   ULONG MirrorFlags;      // (B/B/W  DataBadBits/MirrorActive/SyncNumber)
   ULONG Reserved1;        // (B/B/B/B  BSize/BUMask/BShift/BMask)
   ULONG MirrorStatus;     // (B/B/W DeviceBadBits/DeviceActive/SynchNumber)
   ULONG MirrorTotalSectors;
   ULONG MirrorGroupID;
   ULONG MirrorMemberID[MAX_MIRRORS];
} MIRROR;

typedef struct _FAT_ENTRY
{
   long FATIndex;
   long FATCluster;
} FAT_ENTRY;

typedef struct _SUBALLOC_DIR
{
    ULONG  Flag;
    ULONG  Reserved1;
    BYTE  ID;
    BYTE  SequenceNumber;
    BYTE  Reserved2[2];
    ULONG  SubAllocationList;
    ULONG  StartingFATChain[28];
} SUBALLOC_DIR;

#endif

#endif

